import stat
from abc import abstractmethod
from enum import Enum
import pathlib
from pathlib import PurePosixPath, PureWindowsPath
from typing import (Callable, Generator, Iterable, List, Optional, Tuple,
                    TYPE_CHECKING, Union)


if TYPE_CHECKING:
    from cerulean.file_system_impl import FileSystemImpl


class EntryType(Enum):
    DIRECTORY = 1
    FILE = 2
    SYMBOLIC_LINK = 3
    CHARACTER_DEVICE = 4
    BLOCK_DEVICE = 5
    FIFO = 6
    SOCKET = 7


class Permission(Enum):
    OWNER_READ = stat.S_IRUSR
    OWNER_WRITE = stat.S_IWUSR
    OWNER_EXECUTE = stat.S_IXUSR

    GROUP_READ = stat.S_IRGRP
    GROUP_WRITE = stat.S_IWGRP
    GROUP_EXECUTE = stat.S_IXGRP

    OTHERS_READ = stat.S_IROTH
    OTHERS_WRITE = stat.S_IWOTH
    OTHERS_EXECUTE = stat.S_IXOTH

    SUID = stat.S_ISUID
    SGID = stat.S_ISGID
    STICKY = stat.S_ISVTX


AbstractPath = Union[pathlib.Path, PurePosixPath, PureWindowsPath]


class Path:
    """A path on a file system.

    This class implements the pathlib.PurePosixPath interface \
    fully, and a pathlib.PosixPath-like interface, although it has \
    some omissions, additions, and improvements to make it more \
    compatible with remote and non-standard file systems.

    To make a Path, create a FileSystem first, then use the / operator \
    on it, e.g. fs / 'home' / 'user'. Do not construct objects of this \
    class directly.

    Attributes:
        filesystem: The file system that this path is on.
    """
    def __init__(self, filesystem: 'FileSystemImpl', path: AbstractPath) -> None:
        if isinstance(path, Path):
            raise RuntimeError('AAAAAAARGH!')
        self.__path = path
        self.filesystem = filesystem

    def __str__(self) -> str:
        return str(self.__path)

    # PurePath operators
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Path):
            return NotImplemented
        return (self.filesystem == other.filesystem
                and self.__path == other.__path)

    def __neq__(self, other: object) -> bool:
        if not isinstance(other, Path):
            return NotImplemented
        return not (self == other)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Path):
            return NotImplemented
        if self.filesystem != other.filesystem:
            raise TypeError('\'<\' not supported between different file systems')
        return self.__path < other.__path

    def __gt__(self, other: object) -> bool:
        return other < self

    def __le__(self, other: object) -> bool:
        return self < other or self == other

    def __ge__(self, other: object) -> bool:
        return other < self or self == other

    def __truediv__(self, suffix: Union[str, 'Path']) -> 'Path':
        if isinstance(suffix, Path):
            path = self.__path / suffix.__path
        else:
            path = self.__path / suffix.strip('/')
        return Path(self.filesystem, path)


    # PurePath attributes and functions

    @property
    def parts(self) -> Tuple[str, ...]:
        """A tuple containing the path's components."""
        return self.__path.parts

    @property
    def drive(self) -> str:
        """The drive letter (including the colon), if any."""
        return self.__path.drive

    @property
    def root(self) -> str:
        """A string representing the root of the filesystem."""
        return self.__path.root

    @property
    def anchor(self) -> str:
        """The concatenation of the drive and the root."""
        return self.__path.anchor

    @property
    def parents(self) -> List['Path']:
        """A sequence containing the logical ancestors of the path."""
        def make_path(path: AbstractPath) -> 'Path':
            return Path(self.filesystem, path)

        return list(map(make_path, self.__path.parents))

    @property
    def parent(self) -> 'Path':
        """The logical parent of the path."""
        return Path(self.filesystem, self.__path.parent)

    @property
    def name(self) -> str:
        """The name of the file or directory, excluding parents but \
        including the suffix.
        """
        return self.__path.name

    @property
    def suffix(self) -> str:
        """The file extension of the file or directory, if any."""
        return self.__path.suffix

    @property
    def suffixes(self) -> List[str]:
        """A list of all the extensions in the file name."""
        return self.__path.suffixes

    @property
    def stem(self) -> str:
        """The name of the file or directory, excluding parents and \
        excluding the suffix.
        """
        return self.__path.stem

    def as_posix(self) -> str:
        """Returns the path as a string with forward slashes."""
        return self.__path.as_posix()

    def as_uri(self) -> str:
        """Returns a URI representing the path.

        This is not yet implemented, please file an issue if you need it.
        """
        raise NotImplementedError('Not yet implemented, please file an issue')

    def is_absolute(self) -> bool:
        """Returns whether the path is absolute."""
        return self.__path.is_absolute()

    def is_reserved(self) -> bool:
        """Return whether the path is reserved.

        This can only happen on Windows on a LocalFileSystem.
        """
        return self.__path.is_reserved()

    def joinpath(self, *other: Union[str, 'Path']) -> 'Path':
        """Joins another path or string onto the back of this one.

        Args:
            other: The other path to append to this one.
        """
        def get_path(segment: Union[str, 'Path']) -> str:
            if isinstance(segment, str):
                return segment
            else:
                return str(segment.__path)

        native_others = map(get_path, other)
        return Path(self.filesystem, self.__path.joinpath(*native_others))

    # TODO: match

    def relative_to(self, *other: Union[str, 'Path']) -> 'Path':
        """Returns a version of this path relative to another path.

        Both paths must be on the same file system.

        Args:
            other: The path to use as a reference.
        """
        def get_path(segment: Union[str, 'Path']) -> str:
            if isinstance(segment, str):
                return segment
            else:
                return str(segment.__path)

        native_others = map(get_path, other)
        return Path(self.filesystem, self.__path.relative_to(*native_others))

    def with_name(self, name: str) -> 'Path':
        """Return a new path with the last component set to `name`.

        Args:
            name: The new name to use.
        """
        return Path(self.filesystem, self.__path.with_name(name))

    def with_suffix(self, suffix: str) -> 'Path':
        """Return a new path with the suffix set to `suffix`

        Args:
            suffix: The new suffix to use.
        """
        return Path(self.filesystem, self.__path.with_suffix(suffix))

    # Existence and contents

    def exists(self) -> bool:
        """Returns true iff a filesystem object exists at this path.

        If the path denotes a symlink, returns whether the symlink \
        points to an existing filesystem object, recursively. If the \
        symlink is part of a link loop, returns False.

        Returns:
            True iff the path exists on the filesystem.
        """
        return self.filesystem._exists(self.__path)

    def mkdir(self,
              mode: int = 0o777,
              parents: bool = False,
              exists_ok: bool = False) -> None:
        """Makes a directory with the given access rights.

        If parents is True, makes parent directories as needed. If \
        exists_ok is True, silently ignores if the directory already \
        exists.

        Args:
            mode: A numerical Posix access permissions mode.
            parents: Whether to make parent directories.
            exists_ok: Don't raise if target already exists.
        """
        self.filesystem._mkdir(self.__path, mode, parents, exists_ok)

    def iterdir(self) -> Generator['Path', None, None]:
        """Iterates through a directory's contents.

        Yields:
            Paths of entries in the directory.
        """
        for entry in self.filesystem._iterdir(self.__path):
            yield Path(self.filesystem, entry)

    def walk(self, topdown: bool = True,
             onerror: Optional[Callable[[OSError], None]] = None,
             followlinks: bool = False
             ) -> Generator[Tuple['Path', List[str], List[str]], None, None]:
        """Walks a directory hierarchy recursively.

        This is a version of Python's ``os.walk()`` function adapted to \
        be a little bit more ``pathlib``-like. It walks the directory \
        and its subdirectories, yielding a tuple \
        (dirpath, dirnames, filenames) for each directory. These are \
        the path of the directory, as a :class:`Path`, a list of \
        strings containing the names of the subdirectories inside this \
        directory, and a list of strings containing the names of the \
        non-directories in this directory respectively.

        If ``topdown`` is True, the triple for a directory will be \
        produced before the triples of its subdirectories; if it is \
        False, it will be produced after (pre- and post-order traversal \
        respectively).

        If ``onerror`` is set, the function it is set to will be called \
        if an error occurs, and passed an instance of OSError. The \
        callback can handle the error in some way, or raise it to end \
        the traversal. The OSError object will have an attribute \
        ``filename`` containing the name of the file that triggered \
        the problem as a string.

        If ``followlinks`` is True, this function will recurse into \
        symlinks that point to directories, if it is False, it will \
        silently skip them.

        Yields:
            Tuples (dirpath, dirnames, filenames), as above.
        """
        dirnames = list()  # type: List[str]
        filenames = list()  # type: List[str]
        try:
            for entry in self.iterdir():
                if entry.is_dir():
                    dirnames.append(str(entry.relative_to(self)))
                else:
                    filenames.append(str(entry.relative_to(self)))
        except OSError as err:
            if onerror is not None:
                onerror(err)

        if topdown:
            yield self, dirnames, filenames

        for dirname in dirnames:
            subdir = self / dirname
            if not subdir.is_symlink() or followlinks:
                yield from subdir.walk(topdown, onerror, followlinks)

        if not topdown:
            yield self, dirnames, filenames

    def rmdir(self, recursive: bool = False) -> None:
        """Removes a directory.

        If recursive is True, remove all files and directories inside \
        as well. If recursive is False, the directory must be empty.
        """
        self.filesystem._rmdir(self.__path, recursive)

    def touch(self) -> None:
        """Updates the access and modification times of file.

        If the file does not exist, it will be created, which is often \
        what this function is used for.
        """
        self.filesystem._touch(self.__path)

    def streaming_read(self) -> Generator[bytes, None, None]:
        """Streams data from a file.

        This is a generator function that generates bytes objects \
        containing consecutive chunks of the file.
        """
        return self.filesystem._streaming_read(self.__path)

    def streaming_write(self, data: Iterable[bytes]) -> None:
        """Streams data to a file.

        Creates a new file (overwriting any existing file) at the \
        current path, and writes data to it from the given iterable.

        Args:
            data: An iterable of bytes containing data to be written.
        """
        self.filesystem._streaming_write(self.__path, data)

    def read_bytes(self) -> bytes:
        """Reads file contents as a bytes object.

        Returns:
            The contents of the file.
        """
        data = bytearray()
        for chunk in self.streaming_read():
            data.extend(chunk)
        return bytes(data)

    def read_text(self, encoding: str = 'utf-8') -> str:
        """Reads file contents as a string.

        Assumes UTF-8 encoding.

        Args:
            encoding: The encoding to assume.

        Returns:
            The contents of the file.
        """
        return self.read_bytes().decode(encoding)

    def write_bytes(self, data: bytes) -> None:
        """Writes bytes to the file.

        Overwrites the file if it exists.

        Args:
            data: The data to be written.
        """
        self.streaming_write([data])

    def write_text(self, text: str, encoding: str = 'utf-8') -> None:
        """Writes text to a file.

        Overwrites the file if it exists.

        Args:
            text: The text to be written.
            encoding: The encoding to use.
        """
        self.write_bytes(text.encode(encoding))

    def rename(self, target: 'Path') -> None:
        """Renames a file.

        The new path must be in the same filesystem. If the new path \
        exists, then it will be overwritten.

        Args:
            target: The new path of the file.
        """
        if target.filesystem != self:
            raise RuntimeError('Cannot rename across file systems')
        self.filesystem._rename(self.__path, target.__path)

    def unlink(self) -> None:
        """Removes a file or device node.

        For removing directories, see rmdir().
        """
        self.filesystem._unlink(self.__path)

    # File type and size

    def is_dir(self) -> bool:
        """Returns whether the path is a directory.

        Returns:
            True iff the path exists and is a directory, or a symbolic \
            link pointing to a directory.
        """
        return self.filesystem._is_dir(self.__path)

    def is_file(self) -> bool:
        """Returns whether the path is a file.

        Returns:
            True iff the path exists and is a file, or a symbolic \
            link pointing to a file.
        """
        return self.filesystem._is_file(self.__path)

    def is_symlink(self) -> bool:
        """Returns whether the path is a symlink.

        Returns:
            True iff the path exists and is a symbolic link.
        """
        return self.filesystem._is_symlink(self.__path)

    def entry_type(self) -> EntryType:
        """Returns the kind of directory entry type the path points to.

        Returns:
            An :class:`EntryType` enum value describing the filesystem \
            entry.
        """
        return self.filesystem._entry_type(self.__path)

    def size(self) -> int:
        """Returns the size of the file.

        Returns:
            An integer with the number of bytes in the file.
        """
        return self.filesystem._size(self.__path)

    # Permissions

    def uid(self) -> Optional[int]:
        """Returns the user id of the owner of the object.

        Returns:
            An integer with the id, or None if not supported.
        """
        return self.filesystem._uid(self.__path)

    def gid(self) -> Optional[int]:
        """Returns the group id associated with the object.

        Returns:
            An integer with the id, or None of not supported.
        """
        return self.filesystem._gid(self.__path)

    def has_permission(self, permission: Permission) -> bool:
        """Checks permissions.

        Args:
            permission: A particular file permission, see Permission

        Returns:
            True iff the object exists and has the given permission.
        """
        return self.filesystem._has_permission(self.__path, permission)

    def set_permission(self, permission: Permission,
                       value: bool = True) -> None:
        """Sets permissions.

        Args:
            permission: The permission to set.
            value: Whether to enable or disable the permission.
        """
        self.filesystem._set_permission(self.__path, permission, value)

    def chmod(self, mode: int) -> None:
        """Sets permissions.

        Args:
            mode: The numerical mode describing the permissions to set. \
                  This uses standard POSIX mode definitions, see \
                  man chmod.
        """
        self.filesystem._chmod(self.__path, mode)

    # Symlinks

    def symlink_to(self, target: 'Path') -> None:
        """Makes a symlink from the current path to the target.

        If this raises an OSError with the message Failed, then the \
        problem may be that the target does not exist.

        Args:
            target: The path to symlink to.

        Raises:
            FileExistsError: if you try to overwrite an existing entry \
                    with a symlink.
        """
        if self.filesystem != target.filesystem:
            raise RuntimeError('Cannot symlink across filesystems')
        self.filesystem._symlink_to(self.__path, target.__path)

    def readlink(self, recursive: bool = False) -> 'Path':
        """Reads the target of a symbolic link.

        Note that the result may be a relative path, which should then \
        be taken relative to the directory containing the link.

        If recursive is True, this function will follow a chain of \
        symlinks until it reaches something that is not a symlink, or \
        until the maximum recursion depth is reached and a \
        RunTimeError is raised.

        Args:
            recursive: Whether to resolve recursively.

        Returns:
            The path that the symlink points to.

        Raises:
            RunTimeError: The recursion depth was reached, probably as a \
                    result of a link loop.
        """
        return self.filesystem._readlink(self.__path, recursive)



from abc import ABC, abstractmethod
from types import TracebackType
from typing import List, Optional, Tuple

from cerulean.util import BaseExceptionType


class Terminal(ABC):
    """Interface for Terminals.

    This is a generic interface class that all terminals inherit \
    from, so you can use it wherever any terminal will do.

    In order to do something useful, you'll want an actual terminal,
    like a :class:`LocalTerminal` or an :class:`SshTerminal`.

    Terminals may hold resources, so you should either use them \
    with a ``with`` statement, or call :meth:`close` on them \
    when you are done with them.
    """

    def __enter__(self) -> 'Terminal':
        return self

    def __exit__(self, exc_type: Optional[BaseExceptionType],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        pass

    def close(self) -> None:
        """Close the terminal.

        This closes any connections and frees resources associated \
        with the terminal. :class:`LocalTerminal` does not require \
        this, but terminals that connect to remote machines do. You \
        may want to always either close a Terminal, or use it as a \
        context manager, to avoid problems if you ever change from \
        a local terminal to a remote one.
        """
        pass

    @abstractmethod
    def run(self,
            timeout: float,
            command: str,
            args: List[str],
            stdin_data: str = None,
            workdir: str = None) -> Tuple[Optional[int], str, str]:
        """Run a shell command.

        Args:
            timeout: How long to wait for the result(s)
            command: The command to run.
            args: A list of arguments to pass
            stdin_data: Data to pass to standard input
            workdir: Working directory to execute in

        Returns:
            A tuple containing the exit code, standard output, and \
            standard error output.
        """
        pass



from abc import ABC, abstractmethod
from time import perf_counter, sleep
from typing import Optional

from cerulean.job_description import JobDescription
from cerulean.job_status import JobStatus


class Scheduler(ABC):
    """Interface for job schedulers.

    To run jobs using a scheduler, you will want to use \
    :class:`SlurmScheduler` or :class:`TorqueScheduler`.
    """

    @abstractmethod
    def submit(self, job_description: JobDescription) -> str:
        """Submit a job for execution.

        Args:
            job_description: A description of the job to run.

        Returns:
            A job id that can be used to keep track of it.
        """
        pass

    @abstractmethod
    def get_status(self, job_id: str) -> JobStatus:
        """Look up the status of a job.

        This method is used to check if a job is still in the queue, \
        running, or done.

        Args:
            job_id: A job id string obtained from :meth:`submit`.

        Returns:
            The status of the job as a :class:`JobStatus`
        """
        pass

    @abstractmethod
    def get_exit_code(self, job_id: str) -> Optional[int]:
        """Get the exit code of a finished job.

        Once a job is done, its exit code may be requested using this \
        method. If the job is still running or failed to start, then \
        there is no exit code, and None will be returned.

        Args:
            job_id: A job id string obtained from :meth:`submit`.

        Returns:
            The exit code, or None if there is none.
        """
        pass

    @abstractmethod
    def cancel(self, job_id: str) -> None:
        """Cancel a running job.

        Submits a cancellation request for a job to the scheduler.

        Args:
            job_id: Id of the job to be cancelled.
        """
        pass

    def wait(self, job_id: str, time_out: float = -1.0) -> Optional[int]:
        """Wait until the job is done.

        Will wait approximately time_out seconds for the job to finish.
        Returns the exit code if the job finished, otherwise None.
        Resolution is 0.1s, or time_out / 20, whichever is larger.

        Args:
            job_id: The job to wait for.
            time_out: Time to wait in seconds. If negative, wait \
                    forever.

        Returns:
            The exit code of the job.
        """
        if time_out < 0.0:
            time_end = None
        else:
            time_end = perf_counter() + time_out

        sleep_time = max(0.1, time_out / 20.0)

        status = self.get_status(job_id)
        while status != JobStatus.DONE:
            sleep(sleep_time)
            status = self.get_status(job_id)

            if time_end is not None and time_end < perf_counter():
                break

        return self.get_exit_code(job_id)



import logging
import time
from typing import Optional, Union

from cerulean.job_description import JobDescription
from cerulean.job_status import JobStatus
from cerulean.scheduler import Scheduler
from cerulean.terminal import Terminal


logger = logging.getLogger(__name__)


class SlurmScheduler(Scheduler):
    """Represents a Slurm scheduler.

    This class represents a Slurm scheduler, to which it talks through \
    a :class:`Terminal`.

    Arguments:
        terminal: The terminal to use to talk to the scheduler.
    """
    def __init__(self, terminal: Terminal) -> None:
        self.__terminal = terminal
        exit_code, output, error = self.__terminal.run(10, 'sbatch',
                                                       ['--version'])
        logger.debug('sbatch --version exit code: {}'.format(exit_code))
        logger.debug('sbatch --version output: {}'.format(output))
        logger.debug('sbatch --version error: {}'.format(error))

    def submit(self, job_description: JobDescription) -> str:
        if job_description.command is None:
            raise ValueError('Job description is missing a command')

        job_script = _job_desc_to_job_script(job_description)
        logger.debug('Submitting job script: {}'.format(job_script))

        exit_code, output, error = self.__terminal.run(10, 'sbatch', [],
                                                       job_script, None)

        logger.debug('sbatch exit code: {}'.format(exit_code))
        logger.debug('sbatch output: {}'.format(output))
        logger.debug('sbatch error: {}'.format(error))
        if exit_code != 0:
            raise RuntimeError('Error running sbatch: {}'.format(error))

        job_id = output.strip().split(' ')[-1]
        return job_id

    def get_status(self, job_id: str) -> JobStatus:
        logger.debug('Calling squeue -j {} -h -o %T'.format(job_id))
        exit_code, output, error = self.__terminal.run(
            10, 'squeue', ['-j', job_id, '-h', '-o', '%T'], None, None)
        logger.debug('squeue exit code: {}'.format(exit_code))
        logger.debug('squeue output: {}'.format(output))
        logger.debug('squeue error: {}'.format(error))

        status = output.strip()
        if status == '':
            # Seems like SLURM sometimes does not show the job, possibly
            # because it's transitioning to COMPLETING. So
            # if we don't find it, try again to be a bit more robust.
            time.sleep(2.0)
            logger.debug('No answer from Slurm, trying again...')
            exit_code, output, error = self.__terminal.run(
                10, 'squeue', ['-j', job_id, '-h', '-o', '%T'])
            status = output.strip()
            logger.debug('squeue output 2: {}'.format(status))

        status_map = {
            'PENDING': JobStatus.WAITING,
            'CONFIGURING': JobStatus.WAITING,
            'RUNNING': JobStatus.RUNNING,
            'SUSPENDED': JobStatus.RUNNING,
            'COMPLETING': JobStatus.RUNNING,
            'BOOT_FAIL': JobStatus.DONE,
            'CANCELLED': JobStatus.DONE,
            'COMPLETED': JobStatus.DONE,
            'FAILED': JobStatus.DONE,
            'TIMEOUT': JobStatus.DONE,
            'PREEMPTED': JobStatus.DONE,
            'NODE_FAIL': JobStatus.DONE,
            'REVOKED': JobStatus.DONE,
            'SPECIAL_EXIT': JobStatus.DONE
        }
        try:
            job_status = status_map[status]
        except KeyError:
            job_status = JobStatus.DONE

        logger.debug('get_status returning {}'.format(job_status.name))

        return job_status

    def get_exit_code(self, job_id: str) -> Optional[int]:
        if self.get_status(job_id) != JobStatus.DONE:
            return None

        logger.debug('get_exit_code() running sacct -j {} --noheader'
                     ' --format=ExitCode'.format(job_id))
        err, output, error = self.__terminal.run(
            10, 'sacct', ['-j', job_id, '--noheader', '--format=ExitCode'])
        logger.debug('sacct exit code: {}'.format(err))
        logger.debug('sacct output: {}'.format(output))
        logger.debug('sacct error: {}'.format(error))
        if err != 0:
            raise RuntimeError('Error running sacct: {}'.format(error))

        if output.lstrip() == '':
            return None
        exit_code = int(output.lstrip().split(':')[0])
        return exit_code

    def cancel(self, job_id: str) -> None:
        logger.debug('Running scancel {}'.format(job_id))
        self.__terminal.run(10, 'scancel', [job_id])


def _job_desc_to_job_script(job_description: JobDescription) -> str:
    job_script = '#!/bin/bash\n'
    job_script = _add_option(job_script, 'job-name', job_description.name)

    if job_description.time_reserved is not None:
        job_script += '#SBATCH --time={}\n'.format(
            _seconds_to_time(job_description.time_reserved))

    job_script = _add_option(job_script, 'partition',
                             job_description.queue_name)
    job_script = _add_option(job_script, 'nodes', job_description.num_nodes)
    job_script = _add_option(job_script, 'ntasks-per-node',
                             job_description.mpi_processes_per_node)
    if job_description.mpi_processes_per_node is not None:
        job_script = _add_option(job_script, 'overcommit', '')

    if job_description.extra_scheduler_options is not None:
        job_script += '#SBATCH {}\n'.format(
                job_description.extra_scheduler_options)

    for name, value in job_description.environment.items():
        job_script += "export {}='{}'\n".format(name, value)

    if job_description.working_directory is not None:
        job_script += 'cd {}\n'.format(job_description.working_directory)

    job_script += '{}'.format(job_description.command)
    job_script += ' {}'.format(' '.join(job_description.arguments))

    if job_description.stdout_file is not None:
        job_script += ' >{}'.format(job_description.stdout_file)

    if job_description.stderr_file is not None:
        job_script += ' 2>{}'.format(job_description.stderr_file)
    return job_script


def _add_option(job_script: str, option: str,
                value: Optional[Union[int, str]]) -> str:
    if value is not None:
        if value is not '':
            return job_script + '#SBATCH --{}={}\n'.format(option, value)
        else:
            return job_script + '#SBATCH --{}\n'.format(option)
    else:
        return job_script


def _seconds_to_time(seconds: int) -> str:
    """Converts seconds to a SLURM allocation duration.

    Args:
        seconds: The number of seconds to reserve.

    Returns:
        A string of the form DD:HH:MM:SS.
    """
    seconds_per_day = 60 * 60 * 24
    seconds_per_hour = 60 * 60
    seconds_per_minute = 60

    days = seconds // seconds_per_day
    seconds -= days * seconds_per_day

    hours = seconds // seconds_per_hour
    seconds -= hours * seconds_per_hour

    minutes = seconds // seconds_per_minute
    seconds -= minutes * seconds_per_minute
    return '{:02d}-{:02d}:{:02d}:{:02d}'.format(days, hours, minutes, seconds)



import logging
from time import perf_counter
from typing import Callable, Generator, Iterable, Optional

from cerulean.path import Path, Permission


CopyCallback = Callable[[int, int], None]
"""The type of a callback function for the copy() function.

A callback function takes two arguments, the number of bytes copied \
thus far, and the approximate total number of bytes to copy. To \
interrupt the copy operation, raise an exception.
"""


def copy(source_path: Path,
         target_path: Path,
         overwrite: str = 'never',
         copy_into: bool = True,
         copy_permissions: bool = False,
         callback: Optional[CopyCallback] = None) -> None:
    """Copy a file or directory from one path to another.

    Note that source_path and target_path may be paths on different \
    file systems.

    The overwrite parameter decides what to do if a file is encountered \
    on the target side that would be overwritten. If overwrite equals \
    'raise', then a FileExistsError is raised. If overwrite equals \
    'always', then the file is overwritten (removed and replaced \
    if needed). If overwrite equals 'never', then the existing \
    file is kept, and the source file not copied.

    If the target is a directory and copy_into is True (the default), \
    the source will be copied into the \
    target directory. If an entry with the same name already exists \
    within the target directory, then the overwrite parameter decides \
    what happens. If copy_into is False, the source will \
    be copied on top of the directory, subject to the setting for \
    overwrite.

    If copy_permissions is True, this function will make the target's \
    permissions match those of the source, `including` SETUID, SETGID \
    and sticky bits. If copy_permissions is False, the target's \
    permissions are left at their default values (according to the \
    umask, on Unix-like systems), less any permissions that the source \
    file does not have.

    If callback is provided, it should be a function taking two \
    arguments, the current count of bytes copied and the total number \
    of bytes to be copied. It will be called once at the beginning of \
    the copy operation (with count == 0), once at the end (with count \
    == total), and in between about once per second, if the copy takes \
    long enough. Note that the total number of bytes passed to the \
    callback is approximate, and that the count may be larger than \
    the total if the estimate was off. To abort the copy, raise an \
    exception from the callback function.

    Args:
        source_path: The path to the source file.
        target_path: A path to copy it to.
        overwrite: Selects behaviour when the target exists.
        copy_into: Whether to copy into target directories.
        copy_permissions: Whether to copy permissions along.
        callback: A callback function to call regularly with progress \
                reports.
    """
    if overwrite not in ['always', 'never', 'raise']:
        raise ValueError('Invalid value for overwrite. Valid values are' +
                         '"always", "never" and "raise".')

    if target_path.is_dir() and copy_into:
        target_path = target_path / source_path.name

    size = _get_approx_size(source_path)

    if callback is not None:
        callback(0, size)

    total_written = _copy(source_path, target_path, overwrite,
                          copy_permissions, source_path, callback, 0, size)

    if callback is not None:
        callback(total_written, size)


def _copy(source_path: Path, target_path: Path, overwrite: str,
          copy_permissions: bool, context: Optional[Path],
          callback: Optional[CopyCallback], already_written: int, size: int
          ) -> int:
    """Copy a file or directory from one path to another.

    See the documentation of copy() for the required behaviour.

    The context path is guaranteed to be a prefix of source_path, on \
    the same file system.

    Args:
        source_path: The path to the source file.
        target_path: A path to copy it to.
        overwrite: Selects behaviour when the target exists.
        copy_permissions: Whether to copy permissions along.
        context: Root of the tree we are copying, or None.
        callback: A callback function to call.
        already_written: Starting count of bytes written.
        size: Approximate total size of data to copy.

    Returns:
        The approximate total number of bytes written.
    """
    logging.debug('Copying {} to {}'.format(source_path, target_path))
    target_path_exists = target_path.exists() or target_path.is_symlink()
    if source_path.is_symlink():
        if _copy_symlink(source_path, target_path, overwrite, context):
            return already_written
    if source_path.is_file():
        already_written = _copy_file(source_path, target_path, overwrite,
                                     copy_permissions, callback,
                                     already_written, size)
    elif source_path.is_dir():
        already_written = _copy_dir(source_path, target_path, overwrite,
                                    copy_permissions, context, callback,
                                    already_written, size)
    else:
        # We don't copy special entries or broken links
        logging.debug(
            'Skipping special entry or broken link {}'.format(source_path))
    return already_written


def _copy_symlink(source_path: Path, target_path: Path, overwrite: str,
                  context: Optional[Path]) -> bool:
    """Copy a symlink.

    If overwrite is True, overwrites the target. Copies links within \
    the context as links and returns True, otherwise returns False.
    """
    target_path_exists = target_path.exists() or target_path.is_symlink()
    if not target_path_exists or overwrite == 'always':
        if context is not None:
            linked_path = source_path.readlink(recursive=False)
            if context in linked_path.parents:
                rel_path = linked_path.relative_to(context)
                logging.debug('Making relative link from {} to {}'.format(
                    target_path, rel_path))
                target_fs = target_path.filesystem
                if target_path.exists() or target_path.is_symlink():
                    if target_path.is_dir():
                        target_path.rmdir(recursive=True)
                    else:
                        target_path.unlink()
                target_path.symlink_to(target_fs / str(rel_path))
                return True
        return False  # fall through and copy as file or directory
    elif overwrite == 'raise':
        raise FileExistsError('Target path exists, not overwriting')
    return True  # target path exists, fail silently


def _copy_file(source_path: Path, target_path: Path, overwrite: str,
               copy_permissions: bool, callback: Optional[CopyCallback],
               already_written: int, size: int) -> int:
    """Copy a file.

    Returns the number of bytes written.
    """
    target_path_exists = target_path.exists() or target_path.is_symlink()

    if not target_path_exists or overwrite == 'always':
        logging.debug('Copying file from {} to {}'.format(
                source_path, target_path))

        if not target_path.is_symlink() and target_path.is_dir():
            target_path.rmdir(recursive=True)
        elif target_path.exists():
            target_path.unlink()
        target_path.touch()

        perms = dict()
        for permission in Permission:
            perms[permission] = target_path.has_permission(permission)

        target_path.chmod(0o600)
        target_path.streaming_write(
                _call_back(callback, perf_counter() + 1.0, already_written,
                           size, source_path.streaming_read()))

        already_written += source_path.size()

        for permission in Permission:
            if copy_permissions:
                target_path.set_permission(
                        permission, source_path.has_permission(permission))
            else:
                target_path.set_permission(
                        permission,
                        perms[permission]
                        and source_path.has_permission(permission))

    elif overwrite == 'raise':
        raise FileExistsError('Target path exists, not overwriting')

    return already_written


def _call_back(callback: Optional[CopyCallback], next_callback: float,
               already_written: int, total_size: int, stream: Iterable[bytes]
               ) -> Generator[bytes, None, None]:
    """Calls the callback every second or so.
    """
    written_here = 0
    for chunk in stream:
        yield chunk
        written_here += len(chunk)
        if perf_counter() >= next_callback:
            if callback is not None:
                callback(already_written + written_here, total_size)
            next_callback = perf_counter() + 1.0


def _copy_dir(source_path: Path, target_path: Path, overwrite: str,
              copy_permissions: bool, context: Optional[Path],
              callback: Optional[CopyCallback], already_written: int, size: int
              ) -> int:
    """Copy a directory recursively.
    """
    target_path_exists = target_path.exists() or target_path.is_symlink()

    if target_path_exists:
        if overwrite == 'always':
            if not target_path.is_dir():
                target_path.unlink()
        elif overwrite == 'raise':
            raise FileExistsError('Target path exists, not overwriting')
        elif overwrite == 'never':
            return already_written

    if not target_path.exists():
        logging.debug('Making new dir {}'.format(target_path))
        target_path.mkdir()

    perms = dict()
    for permission in Permission:
        perms[permission] = target_path.has_permission(permission)

    target_path.chmod(0o700)

    for entry in source_path.iterdir():
        logging.debug('Recursively copying entry {}'.format(entry))
        already_written = _copy(entry, target_path / entry.name, overwrite,
                                copy_permissions, context, callback,
                                already_written, size)

    for permission in Permission:
        if copy_permissions:
            target_path.set_permission(
                    permission, source_path.has_permission(permission))
        else:
            target_path.set_permission(
                    permission,
                    perms[permission]
                    and source_path.has_permission(permission))
    return already_written


def _get_approx_size(path: Path) -> int:
    count = 0
    if not path.is_symlink() and path.is_file():
        count += path.size()
    elif not path.is_symlink() and path.is_dir():
        for subdir in path.iterdir():
            count += _get_approx_size(subdir)
    return count



import logging
import socket
from time import perf_counter
from types import TracebackType
from typing import List, Optional, Tuple, Type, TYPE_CHECKING

import paramiko
from cerulean.credential import (Credential, PasswordCredential,
                                 PubKeyCredential)
from cerulean.terminal import Terminal
from cerulean.util import BaseExceptionType


logger = logging.getLogger(__name__)


class SshTerminal(Terminal):
    """A terminal that runs commands over SSH.

    This terminal connects to a host using SSH, then lets you run \
    commands there.

    Arguments:
        host: The hostname to connect to.
        port: The port to connect on.
        credential: The credential to authenticate with.
    """
    def __init__(self, host: str, port: int, credential: Credential) -> None:
        self.__host = host
        self.__port = port
        self.__credential = credential

        self.__transport = self.__ensure_connection(None)
        self.__transport2 = None

    def __enter__(self) -> 'SshTerminal':
        return self

    def __exit__(self, exc_type: Optional[BaseExceptionType],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        self.close()

    def close(self) -> None:
        """Close the terminal.

        This closes any connections and frees resources associated \
        with the terminal.
        """
        self.__transport.close()
        logger.debug('Disconnected from SSH server')

    def _get_sftp_client(self) -> paramiko.SFTPClient:
        """Get an SFTP client using this terminal.

        This function is used by SftpFileSystem to get an SFTP client \
        using this Terminal's connection. This is a private function, \
        but SftpFileSystem is a friend class.

        Returns:
            An SFTP client object using this terminal's connection.
        """
        self.__transport = self.__ensure_connection(self.__transport)
        return paramiko.SFTPClient.from_transport(self.__transport)

    def _get_downstream_sftp_client(self) -> paramiko.SFTPClient:
        """Gets a second SFTP client using this terminal.

        This is a work-around for an issue in paramiko that keeps us \
        from copying data upstream and downstream simultaneously \
        through a single connection with reasonable performance. \
        We solve it by opening a second connection for the downstream \
        part.

        Returns:
            An SFTP client object using a second connection.
        """
        self.__transport2 = self.__ensure_connection(self.__transport2)
        return paramiko.SFTPClient.from_transport(self.__transport2)

    def run(self,
            timeout: float,
            command: str,
            args: List[str],
            stdin_data: str = None,
            workdir: str = None) -> Tuple[Optional[int], str, str]:

        if workdir:
            cmd_str = 'cd {}; {} {}'.format(workdir, command, ' '.join(args))
        else:
            cmd_str = '{} {}'.format(command, ' '.join(args))

        logger.debug('Executing {}'.format(cmd_str))
        last_exception = None  # type: Optional[BaseException]
        start_time = perf_counter()
        while perf_counter() < start_time + timeout:
            self.__transport = self.__ensure_connection(self.__transport)
            try:
                session = self.__transport.open_session()
                logger.debug('Opened session')
                session.exec_command(command=cmd_str)
                logger.debug('exec_command done')
                if stdin_data is not None:
                    session.sendall(bytes(stdin_data, 'utf-8'))
                    session.shutdown_write()
                logger.debug('stdin sent')

                got_all_stdout, stdout_text = self.__get_data_from_channel(
                    session, 'stdout', timeout)
                got_all_stderr, stderr_text = self.__get_data_from_channel(
                    session, 'stderr', timeout)
                logger.debug('got output {} {} {} {}'.format(
                        got_all_stdout, stdout_text, got_all_stderr, stderr_text))
                if not got_all_stdout or not got_all_stderr:
                    logger.debug('Command did not finish within timeout')
                    session.close()
                    return None, stdout_text, stderr_text

                session.settimeout(2.0)
                exit_status = session.recv_exit_status()
                logger.debug('received exit status {}'.format(exit_status))
                session.close()

                if exit_status == -1:
                    raise EOFError('Execution failed, connection'
                                   ' or server issue?')

                logger.debug('Command executed successfully')
                return exit_status, stdout_text, stderr_text
            except paramiko.SSHException as e:
                last_exception = e
            except EOFError as e:
                last_exception = e
            except ConnectionError as e:
                last_exception = e

        raise ConnectionError(str(last_exception))

    def __get_data_from_channel(self, channel: paramiko.Channel,
                                stream_name: str,
                                timeout: float) -> Tuple[bool, str]:
        """Reads text from standard output or standard error."""
        if stream_name == 'stdout':
            receive = paramiko.Channel.recv
        else:
            receive = paramiko.Channel.recv_stderr

        channel.settimeout(timeout)

        data = bytearray()
        try:
            new_data = receive(channel, 1024 * 1024)
            while len(new_data) > 0:
                data.extend(new_data)
                new_data = receive(channel, 1024 * 1024)
        except socket.timeout:
            return False, data.decode('utf-8')

        return True, data.decode('utf-8')

    def __get_key_from_file(self, filename: str,
                            passphrase: Optional[str]) -> paramiko.pkey.PKey:
        key = None
        messages = ''
        try:
            key = paramiko.ed25519key.Ed25519Key.from_private_key_file(
                filename=filename, password=passphrase)
        except paramiko.ssh_exception.SSHException as e:
            key = None
            messages += '{}; '.format(e)

        if key is None:
            try:
                key = paramiko.ecdsakey.ECDSAKey.from_private_key_file(
                    filename=filename, password=passphrase)
            except paramiko.ssh_exception.SSHException as e:
                key = None
                messages += '{}; '.format(e)

        if key is None:
            try:
                key = paramiko.rsakey.RSAKey.from_private_key_file(
                    filename=filename, password=passphrase)
            except paramiko.ssh_exception.SSHException as e:
                key = None
                messages += '{}; '.format(e)

        if key is None:
            logger.debug('Invalid key: {}'.format(messages))
            raise RuntimeError(
                'Invalid key specified, could not open as RSA, ECDSA or Ed25519 key'
            )

        return key

    def __ensure_connection(self, transport: paramiko.Transport
                            ) -> paramiko.Transport:
        if transport is None or not transport.is_active():
            if transport is not None:
                transport.close()
            transport = paramiko.Transport((self.__host, self.__port))
            logger.info('Connecting to {} on port {}'.format(self.__host, self.__port))
            try:
                if isinstance(self.__credential, PasswordCredential):
                    logger.debug('Authenticating using a password')
                    transport.connect(
                        username=self.__credential.username,
                        password=self.__credential.password)
                elif isinstance(self.__credential, PubKeyCredential):
                    logger.debug('Authenticating using a public key')
                    key = self.__get_key_from_file(self.__credential.public_key,
                                                   self.__credential.passphrase)
                    transport.connect(username=self.__credential.username, pkey=key)
                else:
                    raise RuntimeError('Unknown kind of credential')
                logger.info('Connection (re)established')
            except paramiko.SSHException:
                raise ConnectionError('Cerulean was disconnected and could not reconnect')
        return transport



import logging
from subprocess import PIPE, Popen
from typing import List, Optional, Tuple

from cerulean.terminal import Terminal


logger = logging.getLogger(__name__)


class LocalTerminal(Terminal):
    """A Terminal for running commands on the local machine.

    To create one, just do ``term = LocalTerminal()``.
    """
    def run(self,
            timeout: float,
            command: str,
            args: List[str],
            stdin_data: str = None,
            workdir: str = None) -> Tuple[Optional[int], str, str]:

        all_args = [command] + args
        if workdir is not None:
            workdir = str(workdir)
        logger.debug('LocalTerminal running {}'.format(all_args))
        with Popen(
                all_args,
                stdin=PIPE,
                stdout=PIPE,
                cwd=workdir,
                universal_newlines=True) as process:
            stdout_text, stderr_text = process.communicate(
                stdin_data, timeout=timeout)

        logger.debug('LocalTerminal output {}'.format(stdout_text))
        logger.debug('LocalTerminal error {}'.format(stderr_text))
        if stderr_text is None:
            stderr_text = ''
        return process.returncode, stdout_text, stderr_text



from typing import Dict, List, Optional

from pathlib import PurePath


class JobDescription:
    """Describes a job to submit to a scheduler.

    Attributes:
        name (str): The name of the job, with which it will show up in \
                the scheduler's queue. Cerulean does not use the name, \
                but it may be useful if you manually check the queue.
        working_directory (str): The working directory to execute in.
        environment (Dict[str, str]): A dictionary of environment \
                variables to define, and their values.
        command (str): The command to execute.
        arguments (list of str): A list of arguments to pass. If needed, \
                you need to add quotes yourself, the arguments will not \
                be escaped by cerulean.
        stdout_file (str): File to direct standard output to.
        stderr_file (str): File to direct standard error to.
        queue_name (str): Name of the queue to submit to.
        time_reserved (int): Time to reserve, in seconds.
        num_nodes (int): The number of nodes to reserve.
        mpi_processes_per_node (int): Number of MPI processes to start per \
                node.
        extra_scheduler_options (str): Additional options to add to the \
                scheduler command line on job submission. Note that these \
                are scheduler-specific!
    """

    def __init__(self) -> None:
        self.name = 'cerulean'
        self.working_directory = None  # type: Optional[str]
        self.environment = dict()   # type: Dict[str, str]
        self.command = None # type: Optional[str]
        self.arguments = [] # type: List[str]
        self.stdout_file = None # type: Optional[str]
        self.stderr_file = None # type: Optional[str]
        self.queue_name = None  # type: Optional[str]
        self.time_reserved = None   # type: Optional[int]
        self.num_nodes = None   # type: Optional[int]
        self.mpi_processes_per_node = None  # type: Optional[int]
        self.extra_scheduler_options = None  # type: Optional[str]



import logging
import stat
from pathlib import PurePosixPath
from types import TracebackType
from typing import cast, Generator, Iterable, Optional

import paramiko
from cerulean.file_system_impl import FileSystemImpl
from cerulean.path import AbstractPath, EntryType, Path, Permission
from cerulean.ssh_terminal import SshTerminal
from cerulean.util import BaseExceptionType


logger = logging.getLogger(__name__)


class SftpFileSystem(FileSystemImpl):
    """A FileSystem implementation that connects to an SFTP server.

    SftpFileSystem supports the / operation:

    .. code-block:: python

      fs / 'path'

    which produces a :class:`Path`, through which you can do things \
    with the remote files.

    It is also a context manager, so that you can (and should!) use it \
    with a ``with`` statement, which will ensure that the connection \
    is closed when you are done with the it. Alternatively, you can \
    call :meth:`close` to close the connection.

    If `own_term` is True, this class assumes that it owns the terminal \
    you gave it, and that it is responsible for closing it when it's \
    done with it. If you share an SshTerminal between an SftpFileSystem \
    and a scheduler, or use the terminal directly yourself, then you \
    want to use False here, and close the terminal yourself when you \
    don't need it any more.

    Args:
        terminal: The terminal to connect through.
        own_term: Whether to close the terminal when the file system \
                is closed.
    """
    def __init__(self, terminal: SshTerminal, own_term: bool = False) -> None:
        self.__terminal = terminal
        self.__own_term = own_term
        self.__ensure_sftp(True)
        self.__sftp2 = None  # type: paramiko.SFTPClient
        self.__max_tries = 3

    def __enter__(self) -> 'SftpFileSystem':
        return self

    def __exit__(self, exc_type: Optional[BaseExceptionType],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        if self.__own_term:
            self.close()

    def close(self) -> None:
        self.__sftp.close()
        logger.info('Disconnected from SFTP server')

    def __truediv__(self, segment: str) -> Path:
        return Path(self, PurePosixPath('/' + segment.strip('/')))

    def _exists(self, path: AbstractPath) -> bool:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        try:
            self.__sftp.stat(str(lpath))
            return True
        except IOError:
            return False

    def _mkdir(self,
              path: AbstractPath,
              mode: int = 0o777,
              parents: bool = False,
              exists_ok: bool = False) -> None:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        if parents:
            for parent in reversed(lpath.parents):
                if not self._exists(parent):
                    self.__sftp.mkdir(str(parent))
                    # The 0o777 is intentional and matches pathlib and
                    # POSIX mkdir
                    self.__sftp.chmod(str(parent), 0o777)
        if self._exists(lpath):
            if not exists_ok:
                raise FileExistsError(
                    'File {} exists and exists_ok was False'.format(lpath))
            else:
                return

        self.__sftp.mkdir(str(lpath))
        self.__sftp.chmod(str(lpath), mode)

    def _iterdir(self, path: AbstractPath) -> Generator[PurePosixPath, None, None]:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        # Note: we're not using listdir_iter here, because it hangs:
        # https://github.com/paramiko/paramiko/issues/1171
        try:
            for entry in self.__sftp.listdir(str(lpath)):
                yield lpath / entry
        except OSError as e:
            # Paramiko omits the filename, which breaks Path.walk()
            # so add it back in here.
            raise OSError(e.errno, e.strerror, str(lpath))

    def _rmdir(self, path: AbstractPath, recursive: bool = False) -> None:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        if not self._exists(lpath):
            return

        if not self._is_dir(lpath):
            raise RuntimeError("Path must refer to a directory")

        if recursive:
            for entry in self.__sftp.listdir_attr(str(lpath)):
                entry_path = lpath / entry.filename
                if self._is_symlink(entry_path):
                    self.__sftp.unlink(str(entry_path))
                elif self._is_dir(entry_path):
                    self._rmdir(entry_path, True)
                else:
                    self.__sftp.unlink(str(entry_path))

        self.__sftp.rmdir(str(lpath))

    def _touch(self, path: AbstractPath) -> None:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        with self.__sftp.file(str(lpath), 'a'):
            pass

    def _streaming_read(self, path: AbstractPath) -> Generator[bytes, None, None]:
        # Buffer size vs. speed (MB/s) against localhost
        #       up      down        local
        # 8k    33      56          159
        # 16k   52      56          145
        # 24k   66      57          150
        # 32k   24      57          149
        # 2M    24      48
        # scp   120     110
        # cp                        172
        def ensure_sftp2(self: 'SftpFileSystem') -> None:
            if self.__sftp2 is None:
                self.__sftp2 = self.__terminal._get_downstream_sftp_client()
            elif not self.__sftp2.get_channel().get_transport().is_active():
                self.__sftp2 = self.__terminal._get_downstream_sftp_client()

        lpath = cast(PurePosixPath, path)
        tries = 0
        while tries < self.__max_tries:
            ensure_sftp2(self)
            try:
                size = self._size(path)
                with self.__sftp2.file(str(lpath), 'rb') as f:
                    f.prefetch(size)
                    data = f.read(24576)
                    while len(data) > 0:
                        yield data
                        data = f.read(24576)
                return
            except paramiko.SSHException as e:
                if 'Server connection dropped' in str(e):
                    tries += 1
                else:
                    raise e
        raise ConnectionError('Too many connection errors.')

    def _streaming_write(self, path: AbstractPath, data: Iterable[bytes]) -> None:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        tries = 0
        while tries < self.__max_tries:
            try:
                with self.__sftp.file(str(lpath), 'wb') as f:
                    f.set_pipelined(True)
                    for chunk in data:
                        f.write(chunk)
                return
            except paramiko.SSHException as e:
                if 'Server connection dropped' in str(e):
                    tries += 1
                else:
                    raise e
        raise ConnectionError('Too many connection errors.')


    def _rename(self, path: AbstractPath, target: AbstractPath) -> None:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        ltarget = cast(PurePosixPath, target)
        self.__sftp.posix_rename(str(lpath), str(ltarget))

    def _unlink(self, path: AbstractPath) -> None:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        self.__sftp.unlink(str(lpath))

    def _is_dir(self, path: AbstractPath) -> bool:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        try:
            return bool(stat.S_ISDIR(self.__stat(lpath).st_mode))
        except FileNotFoundError:
            return False

    def _is_file(self, path: AbstractPath) -> bool:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        try:
            mode = self.__stat(lpath).st_mode
            return bool(stat.S_ISREG(mode))
        except FileNotFoundError:
            return False

    def _is_symlink(self, path: AbstractPath) -> bool:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        try:
            return bool(stat.S_ISLNK(self.__lstat(lpath).st_mode))
        except FileNotFoundError:
            return False

    def _entry_type(self, path: AbstractPath) -> EntryType:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        mode_to_type = [(stat.S_ISDIR, EntryType.DIRECTORY), (stat.S_ISREG,
                                                              EntryType.FILE),
                        (stat.S_ISLNK, EntryType.SYMBOLIC_LINK),
                        (stat.S_ISCHR,
                         EntryType.CHARACTER_DEVICE), (stat.S_ISBLK,
                                                       EntryType.BLOCK_DEVICE),
                        (stat.S_ISFIFO, EntryType.FIFO), (stat.S_ISSOCK,
                                                          EntryType.SOCKET)]

        mode = self.__lstat(lpath).st_mode
        for predicate, result in mode_to_type:
            if predicate(mode):
                return result
        raise RuntimeError('Object is of unknown type, please report a'
                           'Cerulean bug')

    def _size(self, path: AbstractPath) -> int:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        return self.__stat(lpath).st_size

    def _uid(self, path: AbstractPath) -> int:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        return self.__stat(lpath).st_uid

    def _gid(self, path: AbstractPath) -> int:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        return self.__stat(lpath).st_gid

    def _has_permission(self, path: AbstractPath, permission: Permission) -> bool:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        return bool(self.__stat(lpath).st_mode & permission.value)

    def _set_permission(self,
                       path: AbstractPath,
                       permission: Permission,
                       value: bool = True) -> None:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        mode = self.__stat(lpath).st_mode
        if value:
            mode = mode | permission.value
        else:
            mode = mode & ~permission.value
        self._chmod(lpath, mode)

    def _chmod(self, path: AbstractPath, mode: int) -> None:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        self.__sftp.chmod(str(lpath), mode)

    def _symlink_to(self, path: AbstractPath, target: AbstractPath) -> None:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        ltarget = cast(PurePosixPath, target)
        self.__sftp.symlink(str(ltarget), str(lpath))

    def _readlink(self, path: AbstractPath, recursive: bool) -> Path:
        self.__ensure_sftp()
        lpath = cast(PurePosixPath, path)
        if recursive:
            # SFTP's normalize() raises if there's a link loop or a \
            # non-existing target, which we don't want, so we use \
            # our own algorithm.
            max_iter = 32
            cur_path = lpath
            iter_count = 0
            while self._is_symlink(cur_path) and iter_count < max_iter:
                target = PurePosixPath(self.__sftp.readlink(str(cur_path)))
                if not target.is_absolute():
                    target = cur_path.parent / target
                cur_path = target
                iter_count += 1

            if iter_count == max_iter:
                raise RuntimeError('Too many symbolic links detected')

            target = PurePosixPath(self.__sftp.normalize(str(path)))
        else:
            target = PurePosixPath(self.__sftp.readlink(str(path)))
            if not target.is_absolute():
                target = lpath.parent / target

        return Path(self, target)

    def __lstat(self, path: PurePosixPath) -> paramiko.SFTPAttributes:
        return self.__sftp.lstat(str(path))

    def __stat(self, path: PurePosixPath) -> paramiko.SFTPAttributes:
        return self.__sftp.stat(str(path))

    def __ensure_sftp(self, first: bool = False) -> None:
        if first:
            logger.info('Connecting to SFTP server')
            self.__sftp = self.__terminal._get_sftp_client()
            logger.info('Connected to SFTP server')
        elif not self.__sftp.get_channel().get_transport().is_active():
            logger.info('Reconnecting to SFTP server')
            self.__sftp = self.__terminal._get_sftp_client()
            logger.info('Connected to SFTP server')



from abc import ABC


class Credential(ABC):
    """A credential for connecting to remote machines.

    Credentials don't have much in common other than a username, \
    which is best modelled as a public attribute. So this interface \
    is empty, and only here to provide a generic type to represent \
    any credential in the API.

    Attributes:
        username: The name of the user to connect as.
    """
    pass


class PasswordCredential(Credential):
    """A credential comprising a username and password.

    Attributes:
        username: The name of the user to connect as
        password: The password to authenticate with
    """

    def __init__(self, username: str, password: str) -> None:
        """Create a PasswordCredential.

        Args:
            username: The name of the user to connect as
            password: The password to authenticate with
        """
        self.username = username
        self.password = password


class PubKeyCredential(Credential):
    """A credential using a public/private key pair.

    Attributes:
        username: The name of the user to connect as
        public_key: The (local) path to a key file
        passphrase: The passphrase to decrypt the key with; optional.
    """

    def __init__(self, username: str, public_key: str,
                 passphrase: str = None) -> None:
        """Create a PubKeyCredential.

        Args:
            username: The name of the user to connect as
            pub_key: The (local) path to a key file
            passphrase The passphrase to decrypt the key with; optional.
        """
        self.username = username
        self.public_key = public_key
        self.passphrase = passphrase



from typing import cast, Generator, Optional

from cerulean.credential import Credential
from cerulean.direct_gnu_scheduler import DirectGnuScheduler
from cerulean.file_system import FileSystem
from cerulean.local_file_system import LocalFileSystem
from cerulean.local_terminal import LocalTerminal
from cerulean.scheduler import Scheduler
from cerulean.sftp_file_system import SftpFileSystem
from cerulean.slurm_scheduler import SlurmScheduler
from cerulean.ssh_terminal import SshTerminal
from cerulean.terminal import Terminal
from cerulean.torque_scheduler import TorqueScheduler


def make_file_system(protocol: str, location: Optional[str] = None,
                     credential: Optional[Credential] = None
                     ) -> FileSystem:
    """Make a file system object.

    This is a factory function for FileSystem objects. It will \
    instantiate a FileSystem implementation according to the parameters \
    you give it.

    FileSystems may hold resources, so you should either use this \
    function with a ``with`` statement, or call :meth:`close` on the \
    returned object when you are done with it.

    Args:
        protocol: The protocol to use to connect to the file system. \
                Can be `local` or `sftp`. For `local`, location and \
                credential can be omitted.
        location: The location in the form `hostname` or \
                `hostname:port` to connect to.
        credential: The :class:`Credential` to use to connect with.

    Returns:
        An instance of a FileSystem representing the described file \
                system.
    """
    if protocol == 'local':
        return LocalFileSystem()
    elif protocol == 'sftp':
        term = cast(SshTerminal, make_terminal('ssh', location, credential))
        return SftpFileSystem(term, True)
    else:
        raise ValueError('Unknown protocol, use either local or sftp')


def make_terminal(protocol: str, location: Optional[str] = None,
                  credential: Optional[Credential] = None
                  ) -> Terminal:
    """Make a terminal object.

    This is a factory function for Terminal objects. It will \
    instantiate a Terminal implementation according to the parameters \
    you give it.

    Terminals may hold resources, so you should either use this \
    function with a ``with`` statement, or call :meth:`close` on the \
    returned object when you are done with it.

    Args:
        protocol: The protocol to use to connect to the file system. \
                Can be `local` or `sftp`. For `local`, location and \
                credential can be omitted.
        location: The location in the form `hostname` or \
                `hostname:port` to connect to.
        credential: The :class:`Credential` to use to connect with.

    Returns:
        An instance of a FileSystem representing the described file \
                system.
    """
    if protocol == 'local':
        return LocalTerminal()
    elif protocol == 'ssh':
        if location is None:
            raise ValueError(
                    'The ssh protocol requires a location to connect to')
        if credential is None:
            raise ValueError(
                    'The ssh protocol requires a credential to connect with')

        if ':' in location:
            host = location.split(':')[0]
            port = int(location.split(':')[1])
        else:
            host = location
            port = 22
        return SshTerminal(host, port, credential)
    else:
        raise ValueError('Unknown protocol, use either local or ssh')


def make_scheduler(name: str, terminal: Terminal) -> Scheduler:
    """Make a scheduler object.

    This is a factory function for Scheduler objects. It will \
    instantiate a Scheduler implementation according to the parameters \
    you give it, which talks to the supplied Terminal.

    Args:
        name: The name of the scheduler. One of ``directgnu``,
                ``slurm``, or ``torque``.
        terminal: The terminal this Scheduler will communicate on.

    Returns:
        The Scheduler.
    """
    if name == 'directgnu':
        return DirectGnuScheduler(terminal)
    elif name == 'slurm':
        return SlurmScheduler(terminal)
    elif name == 'torque':
        return TorqueScheduler(terminal)
    else:
        raise ValueError('Unknown scheduler type {} specified, expected one of'
                         ' directgnu, slurm, or torque.'.format(name))



import stat
import traceback
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Generator, Iterable

from cerulean.file_system import FileSystem
from cerulean.path import AbstractPath, EntryType, Path, Permission


class FileSystemImpl(FileSystem):
    """Abstract class for file system implementations.

    This class exists to document the (internal!) interface and to \
    help keep it consistent between the different kinds of file \
    systems, as well as to implement some shared convenience functions.

    You will want to use a  concrete file system, like LocalFileSystem \
    or SftpFileSystem.
    """

    # Existence and contents

    @abstractmethod
    def _exists(self, path: AbstractPath) -> bool:
        pass

    @abstractmethod
    def _mkdir(self,
              path: AbstractPath,
              mode: int = 0o777,
              parents: bool = False,
              exists_ok: bool = False) -> None:
        pass

    @abstractmethod
    def _iterdir(self, path: AbstractPath) -> Generator[AbstractPath, None, None]:
        pass

    @abstractmethod
    def _rmdir(self, path: AbstractPath, recursive: bool = False) -> None:
        pass

    @abstractmethod
    def _touch(self, path: AbstractPath) -> None:
        pass

    @abstractmethod
    def _streaming_read(self, path: AbstractPath) -> Generator[bytes, None, None]:
        pass

    @abstractmethod
    def _streaming_write(self, path: AbstractPath, data: Iterable[bytes]) -> None:
        pass

    # Convenience functions are implemented here, on top of the internal API
    def _read_bytes(self, path: AbstractPath) -> bytes:
        data = bytearray()
        for chunk in self._streaming_read(path):
            data += chunk
        return bytes(data)

    def _read_text(self, path: AbstractPath, encoding: str = 'utf-8',
                  errors: str = 'strict') -> str:
        return self._read_bytes(path).decode(encoding, errors)

    def _write_bytes(self, path: AbstractPath, data: bytes) -> None:
        self._streaming_write(path, [data])

    def _write_text(self,
                   path: AbstractPath,
                   data: str,
                   encoding: str = 'utf-8',
                   errors: str = 'strict') -> None:
        self._write_bytes(path, data.encode(encoding, errors))

    @abstractmethod
    def _rename(self, path: AbstractPath, target: AbstractPath) -> None:
        pass

    @abstractmethod
    def _unlink(self, path: AbstractPath) -> None:
        pass

    # File type and size

    @abstractmethod
    def _is_dir(self, path: AbstractPath) -> bool:
        pass

    @abstractmethod
    def _is_file(self, path: AbstractPath) -> bool:
        pass

    @abstractmethod
    def _is_symlink(self, path: AbstractPath) -> bool:
        pass

    @abstractmethod
    def _entry_type(self, path: AbstractPath) -> EntryType:
        pass

    @abstractmethod
    def _size(self, path: AbstractPath) -> int:
        pass

    # Permissions

    @abstractmethod
    def _uid(self, path: AbstractPath) -> int:
        pass

    @abstractmethod
    def _gid(self, path: AbstractPath) -> int:
        pass

    @abstractmethod
    def _has_permission(self, path: AbstractPath, permission: Permission) -> bool:
        pass

    @abstractmethod
    def _set_permission(self,
                       path: AbstractPath,
                       permission: Permission,
                       value: bool = True) -> None:
        pass

    @abstractmethod
    def _chmod(self, path: AbstractPath, mode: int) -> None:
        pass

    @abstractmethod
    def _symlink_to(self, path: AbstractPath, target: AbstractPath) -> None:
        pass

    @abstractmethod
    def _readlink(self, path: AbstractPath, recursive: bool) -> Path:
        pass



from typing import Type, TYPE_CHECKING

# See https://stackoverflow.com/questions/49959656/typing-exit-in-3-5-fails-on-runtime-but-typechecks
if TYPE_CHECKING:
    BaseExceptionType = Type[BaseException]
else:
    BaseExceptionType = None



import errno
import os
import pathlib
from typing import cast, Any, Generator, Iterable

from cerulean.file_system_impl import FileSystemImpl
from cerulean.path import AbstractPath, EntryType, Path, Permission


class LocalFileSystem(FileSystemImpl):
    """Represents the local file system.

    To create an instance, just call `LocalFileSystem()`.

    LocalFileSystem support a single operation:

    .. code-block:: python

      fs / 'path'

    which produces a :class:`Path`, through which you can do things \
    with local files.

    LocalFileSystem is a context manager, so you can use it in a \
    ``with`` statement, and it has a :meth:`close` method, but since \
    it doesn't hold any resources, you do not need to use them. It may \
    be good to do so anyway, to avoid leaks if you ever replace it with \
    a different :class:`FileSystem` that does.
    """
    def __truediv__(self, segment: str) -> Path:
        # TODO: segment: Union[str, pathlib.Path]?
        absseg = '/' + segment.strip('/')
        path = pathlib.Path(absseg)
        return Path(self, path)

    def _exists(self, path: AbstractPath) -> bool:
        lpath = cast(pathlib.Path, path)
        try:
            return lpath.exists()
        except OSError as e:
            if e.errno == errno.ELOOP:
                return False
            raise

    def _mkdir(self,
              path: AbstractPath,
              mode: int = 0o777,
              parents: bool = False,
              exists_ok: bool = False) -> None:
        lpath = cast(pathlib.Path, path)
        lpath.mkdir(mode, parents, exists_ok)

    def _iterdir(self, path: AbstractPath) -> Generator[AbstractPath, None, None]:
        lpath = cast(pathlib.Path, path)
        for entry in lpath.iterdir():
            yield entry

    def _rmdir(self, path: AbstractPath, recursive: bool = False) -> None:
        lpath = cast(pathlib.Path, path)
        if not lpath.is_dir():
            raise RuntimeError('Path must refer to a directory')

        if recursive:
            for entry in lpath.iterdir():
                if entry.is_symlink():
                    entry.unlink()
                elif entry.is_dir():
                    self._rmdir(entry, True)
                else:
                    entry.unlink()

        lpath.rmdir()

    def _touch(self, path: AbstractPath) -> None:
        lpath = cast(pathlib.Path, path)
        lpath.touch()

    def _streaming_read(self, path: AbstractPath) -> Generator[bytes, None, None]:
        # Buffer size vs. speed (MB/s) against localhost
        #       up      down        local
        # 8k    33      56          159
        # 16k   52      56          145
        # 24k   66      57          150
        # 32k   24      57          149
        # 2M    24      48
        # scp   120     110
        # cp                        172
        lpath = cast(pathlib.Path, path)
        with lpath.open('rb') as f:
            data = f.read(24576)
            while len(data) > 0:
                yield data
                data = f.read(24576)

    def _streaming_write(self, path: AbstractPath, data: Iterable[bytes]) -> None:
        lpath = cast(pathlib.Path, path)
        with lpath.open('wb') as f:
            for chunk in data:
                f.write(chunk)

    def _rename(self, path: AbstractPath, target: AbstractPath) -> None:
        lpath = cast(pathlib.Path, path)
        ltarget = cast(pathlib.Path, target)
        lpath.replace(pathlib.Path(ltarget))

    def _unlink(self, path: AbstractPath) -> None:
        lpath = cast(pathlib.Path, path)
        lpath.unlink()

    def _is_dir(self, path: AbstractPath) -> bool:
        lpath = cast(pathlib.Path, path)
        return lpath.is_dir()

    def _is_file(self, path: AbstractPath) -> bool:
        lpath = cast(pathlib.Path, path)
        return lpath.is_file()

    def _is_symlink(self, path: AbstractPath) -> bool:
        lpath = cast(pathlib.Path, path)
        return lpath.is_symlink()

    def _entry_type(self, path: AbstractPath) -> EntryType:
        lpath = cast(pathlib.Path, path)
        # Note: symlink goes first, because is_dir() and is_file() will
        # dereference and return true, while we want to say it's a
        # symlink and leave it at that.
        pred_to_type = [(pathlib.Path.is_symlink,
                         EntryType.SYMBOLIC_LINK), (pathlib.Path.is_dir,
                                                    EntryType.DIRECTORY),
                        (pathlib.Path.is_file,
                         EntryType.FILE), (pathlib.Path.is_char_device,
                                           EntryType.CHARACTER_DEVICE),
                        (pathlib.Path.is_block_device, EntryType.BLOCK_DEVICE),
                        (pathlib.Path.is_fifo,
                         EntryType.FIFO), (pathlib.Path.is_socket,
                                           EntryType.SOCKET)]

        for pred, entry_type in pred_to_type:
            if pred(lpath):
                return entry_type
        raise RuntimeError('Object is of unknown type, please report a'
                           'Cerulean bug')

    def _size(self, path: AbstractPath) -> int:
        lpath = cast(pathlib.Path, path)
        return lpath.stat().st_size

    def _uid(self, path: AbstractPath) -> int:
        lpath = cast(pathlib.Path, path)
        return lpath.stat().st_uid

    def _gid(self, path: AbstractPath) -> int:
        lpath = cast(pathlib.Path, path)
        return lpath.stat().st_gid

    def _has_permission(self, path: AbstractPath, permission: Permission) -> bool:
        lpath = cast(pathlib.Path, path)
        return bool(lpath.stat().st_mode & permission.value)

    def _set_permission(self, path: AbstractPath, permission: Permission,
                       value: bool = True) -> None:
        lpath = cast(pathlib.Path, path)
        mode = lpath.stat().st_mode
        if value:
            mode = mode | permission.value
        else:
            mode = mode & ~permission.value

        self._chmod(lpath, mode)

    def _chmod(self, path: AbstractPath, mode: int) -> None:
        lpath = cast(pathlib.Path, path)
        lpath.chmod(mode)

    def _symlink_to(self, path: AbstractPath, target: AbstractPath) -> None:
        lpath = cast(pathlib.Path, path)
        ltarget = cast(pathlib.Path, target)
        lpath.symlink_to(ltarget)

    def _readlink(self, path: AbstractPath, recursive: bool) -> Path:
        lpath = cast(pathlib.Path, path)
        if recursive:
            # pathlib.Path.resolve() raises if the link is broken
            # we don't want that, so use our own implementation
            max_iter = 32
            cur_path = lpath
            iter_count = 0
            while cur_path.is_symlink() and iter_count < max_iter:
                target = pathlib.Path(os.readlink(str(cur_path)))
                if not target.is_absolute():
                    target = cur_path.parent / target
                cur_path = target
                iter_count += 1

            if iter_count == max_iter:
                raise RuntimeError('Too many symbolic links detected')

            return Path(self, cur_path)
        else:
            return Path(self, pathlib.Path(
                os.readlink(str(lpath))))



from abc import ABC
from pathlib import PurePath
from types import TracebackType
from typing import cast, TYPE_CHECKING, Optional, Union

from cerulean.path import Path
from cerulean.util import BaseExceptionType


if TYPE_CHECKING:
    from cerulean.file_system_impl import FileSystemImpl


class FileSystem(ABC):
    """Represents a file system.

    This is a generic interface class that all file systems inherit \
    from, so you can use it wherever any file system will do.

    In order to do something useful, you'll want an actual file system,
    like a :class:`LocalFileSystem` or an :class:`SftpFileSystem`.

    FileSystems may hold resources, so you should either use them \
    with a ``with`` statement, or call :meth:`close` on the returned \
    object when you are done with it.

    Beyond that, file systems support a single operation:

    .. code-block:: python

      fs / 'path'

    which produces a :class:`Path`, through which you can do things \
    with files.
    """
    def __enter__(self) -> 'FileSystem':
        return self

    def __exit__(self, exc_type: Optional[BaseExceptionType],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        pass

    def close(self) -> None:
        """Close connections and free resources, if any.

        FileSystem objects may hold resources that need to be freed \
        when you are done with the object. You can free them by calling \
        this function, or you can use the FileSystem as a context \
        manager using a ``with`` statement.
        """
        pass

    def __truediv__(self, segment: str) -> Path:
        raise NotImplementedError()



import logging
from typing import Optional

from cerulean.job_description import JobDescription
from cerulean.job_status import JobStatus
from cerulean.scheduler import Scheduler
from cerulean.terminal import Terminal
from defusedxml import ElementTree


logger = logging.getLogger(__name__)


class TorqueScheduler(Scheduler):
    """Represents a Torque scheduler.

    This class represents a Torque scheduler, to which it talks through \
    a :class:`Terminal`.

    Arguments:
        terminal: The terminal to use to talk to the scheduler.
    """
    def __init__(self, terminal: Terminal) -> None:
        self.__terminal = terminal
        logger.debug('Running qsub --version')
        exit_code, output, error = self.__terminal.run(10, 'qsub',
                                                       ['--version'])
        logger.debug('qsub --version exit_code: {}'.format(exit_code))
        logger.debug('qsub --version std output: {}'.format(output))
        logger.debug('qsub --version std error: {}'.format(error))

    def submit(self, job_description: JobDescription) -> str:
        if job_description.command is None:
            raise ValueError('Job description is missing a command')

        job_script = _job_desc_to_job_script(job_description)
        logger.debug('Running qsub with job script:\n{}'.format(job_script))
        exit_code, output, error = self.__terminal.run(10, 'qsub', ['-'],
                                                       job_script)

        logger.debug('qsub exit code: {}'.format(exit_code))
        logger.debug('qsub std output: {}'.format(output))
        logger.debug('qsub std error: {}'.format(error))

        if exit_code != 0:
            raise RuntimeError('Torque qsub error: {}'.format(error))

        job_id = output.strip().split(' ')[-1]
        return job_id

    def get_status(self, job_id: str) -> JobStatus:
        logger.debug('Running qstat with job id {}'.format(job_id))
        exit_code, output, error = self.__terminal.run(10, 'qstat',
                                                       ['-x', job_id])
        logger.debug('qstat exit code: {}'.format(exit_code))
        logger.debug('qstat output: {}'.format(output))
        logger.debug('qstat error: {}'.format(error))

        if output == '':
            raise RuntimeError(
                    'No output from qstat, could not determine job status.')
        xml_data = ElementTree.fromstring(output)
        if len(xml_data) == 0:
            return JobStatus.DONE

        status = _get_field_from_qstat_xml(xml_data, 'job_state')
        logger.debug('qstat status: {}'.format(status))

        status_map = {
            'W': JobStatus.WAITING,  # waiting for start time
            'Q': JobStatus.WAITING,  # queued
            'H': JobStatus.RUNNING,  # held
            'S': JobStatus.RUNNING,  # suspended
            'R': JobStatus.RUNNING,  # running
            'T': JobStatus.RUNNING,  # being moved
            'E': JobStatus.RUNNING,  # exiting
            'C': JobStatus.DONE  # completed
        }
        try:
            job_status = status_map[status]
        except KeyError:
            raise RuntimeError(
                'Received an unexpected job status {} from qstat'.format(
                    status))

        logger.debug('get_status returning {}'.format(job_status.name))
        return job_status

    def get_exit_code(self, job_id: str) -> Optional[int]:
        if self.get_status(job_id) != JobStatus.DONE:
            return None

        logger.debug('get_exit_code() running qstat -x {}'.format(job_id))
        exit_code, output, error = self.__terminal.run(10, 'qstat',
                                                       ['-x', job_id])
        logger.debug('qstat exit code: {}'.format(exit_code))
        logger.debug('qstat output: {}'.format(output))
        logger.debug('qstat error: {}'.format(error))

        xml_data = ElementTree.fromstring(output)
        job_exit_code = int(_get_field_from_qstat_xml(xml_data, 'exit_status'))
        return job_exit_code

    def cancel(self, job_id: str) -> None:
        logger.debug('cancel() running qdel {}'.format(job_id))
        exit_code, output, error = self.__terminal.run(10, 'qdel', [job_id])
        logger.debug('qdel exit code: {}'.format(exit_code))
        logger.debug('qdel output: {}'.format(output))
        logger.debug('qdel error: {}'.format(error))


def _job_desc_to_job_script(job_description: JobDescription) -> str:
    job_script = '#!/bin/bash\n'
    job_script += '#PBS -N {}\n'.format(job_description.name)

    resources = []
    if job_description.time_reserved is not None:
        resources.append('walltime={}'.format(
            _seconds_to_time(job_description.time_reserved)))

    if job_description.num_nodes is not None:
        resources.append('nodes={}'.format(job_description.num_nodes))

    if job_description.mpi_processes_per_node is not None:
        resources.append('ppn={}'.format(
            job_description.mpi_processes_per_node))

    if len(resources) > 0:
        job_script += '#PBS -l {}\n'.format(':'.join(resources))

    if job_description.queue_name is not None:
        job_script += '#PBS -q {}\n'.format(job_description.queue_name)

    if job_description.extra_scheduler_options is not None:
        job_script += '#PBS {}\n'.format(
                job_description.extra_scheduler_options)

    for name, value in job_description.environment.items():
        job_script += "export {}='{}'\n".format(name, value)

    if job_description.working_directory is not None:
        job_script += 'cd {}\n'.format(job_description.working_directory)

    job_script += '{}'.format(job_description.command)

    args = map(str, job_description.arguments)
    job_script += ' {}'.format(' '.join(args))

    if job_description.stdout_file is not None:
        job_script += ' >{}'.format(job_description.stdout_file)

    if job_description.stderr_file is not None:
        job_script += ' 2>{}'.format(job_description.stderr_file)

    return job_script


def _get_field_from_qstat_xml(xml_data: ElementTree, field_name: str) -> str:
    for job in xml_data:
        for field in job:
            if field.tag == field_name:
                return field.text
    raise RuntimeError(
        'Expected field {} not returned by qstat'.format(field_name))


def _seconds_to_time(seconds: int) -> str:
    """Converts seconds to a Torque allocation duration.

    Args:
        seconds: The number of seconds to reserve.

    Returns:
        A string of the form DD:HH:MM:SS.
    """
    seconds_per_day = 60 * 60 * 24
    seconds_per_hour = 60 * 60
    seconds_per_minute = 60

    days = seconds // seconds_per_day
    seconds -= days * seconds_per_day

    hours = seconds // seconds_per_hour
    seconds -= hours * seconds_per_hour

    minutes = seconds // seconds_per_minute
    seconds -= minutes * seconds_per_minute
    return '{:02d}:{:02d}:{:02d}:{:02d}'.format(days, hours, minutes, seconds)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

setup(
    name='cerulean',
    version='0.3.1',
    description="A Python 3 library for talking to HPC clusters and supercomputers",
    long_description=readme + '\n\n',
    author="Lourens Veen",
    author_email='l.veen@esciencecenter.nl',
    url='https://github.com/MD-Studio/cerulean',
    packages=[
        'cerulean',
    ],
    package_dir={'cerulean':
                 'cerulean'},
    include_package_data=True,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='cerulean',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    install_requires=[
        'defusedxml',
        'paramiko',
        'typing',
        'typing-extensions'
    ],
    setup_requires=[
        # dependency for `python setup.py test`
        'pytest-runner',
        # dependencies for `python setup.py build_sphinx`
        'sphinx',
        'sphinx-rtd-theme',
        'recommonmark'
    ],
    tests_require=[
        'docker',
        'docker-compose',
        'pytest',
        'pytest-cov',
        'pycodestyle',
        'sh',
    ],
    extras_require={
        'dev':  ['prospector[with_pyroma]', 'yapf', 'isort'],
    }
)



"""The :mod:`cerulean` module is the main API for Cerulean.

This module contains all the functions you need to use Cerulean.

Below, you will also find documentation for submodules. That is \
developer documentation, you do not need it to use Cerulean.
"""

__version__ = '0.1.0'

__author__ = 'Lourens Veen'
__email__ = 'l.veen@esciencecenter.nl'


from cerulean.copy_files import copy, CopyCallback
from cerulean.credential import Credential, PasswordCredential, PubKeyCredential
from cerulean.direct_gnu_scheduler import DirectGnuScheduler
from cerulean.factory import make_file_system, make_terminal, make_scheduler
from cerulean.file_system import FileSystem
from cerulean.job_description import JobDescription
from cerulean.job_status import JobStatus
from cerulean.local_file_system import LocalFileSystem
from cerulean.local_terminal import LocalTerminal
from cerulean.path import EntryType, Permission, Path
from cerulean.scheduler import Scheduler
from cerulean.sftp_file_system import SftpFileSystem
from cerulean.slurm_scheduler import SlurmScheduler
from cerulean.ssh_terminal import SshTerminal
from cerulean.terminal import Terminal
from cerulean.torque_scheduler import TorqueScheduler

import logging


logger = logging.getLogger('cerulean')
"""The Cerulean root logger. Use this to set Cerulean's log level.

In particular, if something goes wrong and you want more debug output, you \
can do::

    import logging

    cerulean.logger.setLevel(logging.INFO)

or for even more::

    cerulean.logger.setLevel(logging.DEBUG)
"""

__all__ = ['copy', 'logger', 'make_file_system', 'make_terminal',
           'make_scheduler', 'CopyCallback', 'Credential', 'EntryType',
           'PasswordCredential', 'Permission', 'PubKeyCredential',
           'DirectGnuScheduler', 'FileSystem', 'JobDescription', 'JobStatus',
           'LocalFileSystem', 'LocalTerminal', 'Path', 'Scheduler',
           'SftpFileSystem', 'SlurmScheduler', 'SshTerminal', 'Terminal',
           'TorqueScheduler']



import logging
from typing import Optional

from cerulean.job_description import JobDescription
from cerulean.job_status import JobStatus
from cerulean.scheduler import Scheduler
from cerulean.terminal import Terminal


logger = logging.getLogger(__name__)


class DirectGnuScheduler(Scheduler):
    """A scheduler that runs processes directly on a GNU system.

    This scheduler does not have a queue, instead it launches each \
    job immediately as a process, and uses ps and kill to manage it. \

    Args:
        terminal: The terminal to execute on.
    """

    def __init__(self, terminal: Terminal) -> None:
        self.__terminal = terminal

    def submit(self, job_description: JobDescription) -> str:
        if job_description.command is None:
            raise ValueError('Job description is missing a command')

        if job_description.mpi_processes_per_node is not None:
            raise RuntimeError(
                "mpi_processes_per_node is not supported by DirectGnuScheduler, "
                "because we cannot inject this into the MPI configuration in an environment "
                "without a scheduler. You should call mpirun with an appropriate parameter "
                "instead.")

        job_script = ''
        for name, value in job_description.environment.items():
            job_script += "export {}='{}'\n".format(name, value)
        if job_description.working_directory is not None:
            job_script += 'cd {}\n'.format(job_description.working_directory)
        job_script += 'exit_code_file=$(mktemp)\n'
        # TODO: use ulimit to set a timeout
        job_script += "(\n"
        if job_description.time_reserved is not None:
            job_script += "ulimit -t {}\n".format(
                job_description.time_reserved)
        escaped_command = job_description.command.replace("'", "'\\\''")    # type: ignore
        escaped_args = map(lambda s: s.replace("'", "'\\\''"),
                           job_description.arguments)
        job_script += "bash -c '"
        job_script += '{}'.format(escaped_command)
        job_script += ' {}'.format(' '.join(escaped_args))
        if job_description.stdout_file is not None:
            job_script += ' >{}'.format(job_description.stdout_file)
        if job_description.stderr_file is not None:
            job_script += ' 2>{}'.format(job_description.stderr_file)
        job_script += "' ; "
        job_script += 'echo $? >$exit_code_file'
        job_script += ') >/dev/null 2>/dev/null &\n'
        job_script += 'echo -n $! $exit_code_file\n'
        job_script += 'disown\n'

        logger.debug('Job script: {}'.format(job_script))
        exit_code, output, error = self.__terminal.run(10.0, 'bash', [],
                                                       job_script)

        return output

    def get_status(self, job_id: str) -> JobStatus:
        pid = job_id.split(' ')[0]
        logger.debug('Running ps -p {}'.format(pid))
        exit_code, output, error = self.__terminal.run(10.0, 'ps', ['-p', pid])

        logger.debug('ps exit code: {}'.format(exit_code))
        logger.debug('ps output: {}'.format(output))
        logger.debug('ps error: {}'.format(error))

        if exit_code == 0:
            return JobStatus.RUNNING
        else:
            return JobStatus.DONE

    def get_exit_code(self, job_id: str) -> Optional[int]:
        exit_code_file = job_id.split(' ', maxsplit=1)[1]
        logger.debug('Running cat {}'.format(exit_code_file))
        exit_code, output, error = self.__terminal.run(10.0, 'cat',
                                                       [exit_code_file])
        logger.debug('cat exit code: {}'.format(exit_code))
        logger.debug('cat output: {}'.format(output))
        logger.debug('cat error: {}'.format(error))
        # TODO: delete tempfile?
        try:
            return int(output)
        except ValueError:
            return None

    def cancel(self, job_id: str) -> None:
        pid = job_id.split(' ')[0]
        logger.debug('Running kill {}'.format(pid))
        exit_code, output, error = self.__terminal.run(10.0, 'kill', [pid])

        logger.debug('kill exit code: {}'.format(exit_code))
        logger.debug('kill output: {}'.format(output))
        logger.debug('kill error: {}'.format(error))
        # TODO: Check exit code and return whether it was running?
        # TODO: Check if it's stopped, do a -9 if not?
        pass
